/***************************************************************************
 *   Copyright (C) 2003 by Lorenzo Bettini                                 *
 *   http://www.lorenzobettini.it                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <assert.h>
#include <stdlib.h>

#include "classdecl.h"
#include "multimethods.h"
#include "mainerrormanager.h"

using namespace std;

ClassDecl::~ClassDecl()
{
  if (! base_classes)
    return;

  for (BaseClassesIterator it = base_classes->begin ();
       it != base_classes->end (); ++it)
     delete *it;

  if (multimethods)
    delete multimethods;

  delete base_classes;
}

void
ClassDecl::output(ostream &stream)
{
  stream << getClassStruct() << " " << getClassName ();

  if (base_classes)
    {
      stream << " : ";

      for (BaseClassesIterator it = base_classes->begin ();
           it != base_classes->end (); )
        {
          (*it)->output (stream);
          if ( (++it) != base_classes->end () )
            stream << ", ";
        }
    }

  stream << endl;
  stream << "{" << endl;
}

bool
ClassDecl::isSubClass(const ClassDecl *base_cl, bool equal_ok) const
{
  return (base_cl && isSubClass (base_cl->getClassName (), equal_ok));
}

bool
ClassDecl::isSubClass(const string &base_cl, bool equal_ok) const
{
  if (getClassName () == base_cl)
    return equal_ok;

  if (! base_classes)
    return false;

  const BaseClass *base;

  for (BaseClassesIterator bit = base_classes->begin ();
       bit != base_classes->end (); ++bit)
    {
      base = *bit;
      const string &base_name = base->getClassName();

      if (base->is_template_base_class ())
        {
          // FIXME: handle also template classes
          return false;
        }

      if (! base->getClassDecl ())
        {
          cerr << "undefined class decl for base class "
               << base_name << endl;
          assert (0);
        }

      if (base_name == base_cl ||
          base->getClassDecl()->isSubClass (base_cl))
        return true;
    }

  return false;
}

bool
ClassDecl::unrelated(const ClassDecl *cl) const
{
  return ! isSubClass(cl) && ! cl->isSubClass(this);
}

const MultiMethod *
ClassDecl::retrieveMultiMethod(const string &n) const
{
  if (! multimethods)
    return 0;

  return multimethods->retrieveMultiMethod(n);
}

void
ClassDecl::addMethod(MethDecl *m)
{
  assert (multimethods);

  multimethods->addMethod (m);
}

bool
ClassDecl::hasMethod(const MethDecl *m) const
{
  return (multimethods && multimethods->hasMethod (m));
}

const MethDecl *
ClassDecl::getMethod(const MethDecl *m) const
{
  if (! multimethods)
    return 0;

  return multimethods->getMethod(m);
}

void
ClassDecl::addMultiMethod(MultiMethod *mm)
{
  if (! multimethods)
    multimethods = new MultiMethods;

  if (multimethods->find (mm) != multimethods->end ())
    {
      errormanager->print
       ("duplicate branch set for method " + mm->getMethodName(), mm);
      errormanager->print
       ("previous declaration is here", *(multimethods->find(mm)));
      errormanager->exit_failure ();
    }

  multimethods->insert (mm);
  mm->setClassDecl (this);
}

bool
ClassDecl::contains_multimethods() const
{
  return (multimethods && multimethods->size() > 0);
}

AccessType
ClassDecl::getMethodAccessType(const string &m) const
{
  const MultiMethod *multimethods = retrieveMultiMethod(m);

  assert (multimethods);

  return multimethods->getAccessType();
}

