/***************************************************************************
 *   Copyright (C) 2003 by Lorenzo Bettini                                 *
 *   http://www.lorenzobettini.it                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <sys/time.h>

#include "cmdline.h"
#include "sourceanalyser.h"
#include "sourcemodifierdouble.h"
#include "sourcemodifiermodular.h"
#include "branchanalyser.h"
#include "fileutil.h"
#include "sourcestructs.h"
#include "sourcegenerator.h"
#include "includemodifier.h"
#include "multimethodtypes.h"
#include "dbtypegenerator.h"
#include "inheritedtypegenerator.h"
#include "parambinder.h"

#include "mainprogelems.h"
#include "mainglobalvars.h"
#include "mainverbose.h"
#include "mainerrormanager.h"

using namespace std;

ProgElems *prog_elems;
Verbose *verbose;
ErrorManager *errormanager;

int main(int argc, char *argv[])
{
  string file_name;

  gengetopt_args_info args_info;

  // for statistics
  struct timeval tv1, tv2;
  unsigned int num_of_analysed_files = 0;
  unsigned int num_of_analysed_classes = 0;

  if(cmdline_parser(argc, argv, &args_info) != 0)
    // calls cmdline parser. The user gived bag args if it doesn't return -1
    return 1;

  verbose = new Verbose (args_info.verbose_given != 0);
  errormanager = new ErrorManager;

  input_header_file_ext = args_info.input_header_ext_arg;
  output_header_file_ext = args_info.output_header_ext_arg;
  output_source_file_ext = args_info.output_source_ext_arg;
  output_header_file_suff =
    (args_info.output_header_suff_arg ? args_info.output_header_suff_arg : "");
  output_source_file_suff =
    (args_info.output_source_suff_arg ? args_info.output_source_suff_arg : "");

  renamed_suffix = args_info.rename_suffix_arg;
  renamed_main_suffix = args_info.rename_db_suffix_arg;
  rename_overloaded = args_info.rename_overloaded_flag;
  force_regeneration = (args_info.force_given != 0);
  invade_target = (args_info.invade_target_given != 0);
  test_mode = (args_info.test_mode_given != 0);
  modular = (args_info.modular_given != 0);
  clean_targets = (args_info.clean_targets_given != 0);
  no_linenum = (args_info.no_linenum_given != 0);
  dont_infer_headers = (args_info.dont_infer_headers_given != 0);
  assume_virtual = (args_info.assume_virtual_given != 0);
  assume_multimeth_virtual = (args_info.assume_multimeth_virtual_given != 0 || assume_virtual);
  assume_branch_virtual = (args_info.assume_branch_virtual_given != 0 || assume_virtual);

  if (clean_targets && ! modular)
    {
      cerr << "--clean-targets requires --modular" << endl;
      return EXIT_FAILURE;
    }

  if (args_info.stats_given)
  {
    gettimeofday(&tv1, 0);
  }

  if (args_info.sourcepath_given)
    {
      for (unsigned int i = 0; i < args_info.sourcepath_given; ++i)
        sourcepath.push_back(args_info.sourcepath_arg[i]);
    }

  if (args_info.excludeheaders_given)
    {
      for (unsigned int i = 0; i < args_info.excludeheaders_given; ++i)
        excludeheaders.insert(args_info.excludeheaders_arg[i]);
    }

  if (args_info.input_given)
    {
      SourceStructs sourcestructs;
      SourceStructs modified_declarations;
      SourceStructs modified_methods;
      MultiMethodTypes multi_method_types;
      SourceAnalyser sourceanalyser (&sourcestructs);
      main_path = get_file_path (args_info.input_arg);
      verbose->print ("main path: " + main_path);
      int res = sourceanalyser.analyse_program (args_info.input_arg);
      num_of_analysed_files = sourceanalyser.getAnalysedFileNum();
      num_of_analysed_classes = sourceanalyser.getAnalysedClassNum();

      if (res == EXIT_SUCCESS)
        {
          BranchAnalyser branchanalyser (&sourcestructs, multi_method_types);
          ParamBinder paramBinder (&sourcestructs);

          if (paramBinder.analyse() == EXIT_SUCCESS &&
              branchanalyser.analyse () == EXIT_SUCCESS)
            {
              verbose->print("Multi method types:");
              verbose->print(multi_method_types.toString());

              MultiMethodTypes inherited_types;
              InheritedTypeGenerator inheritedtypegenerator
                  (multi_method_types, inherited_types);
              inheritedtypegenerator.generate();
              verbose->print ("Inherited types:");
              verbose->print (inherited_types.toString());

              if (inheritedtypegenerator.found_error())
                {
                  errormanager->exit_failure();
                }

              MultiMethodTypes top_types;
              TopTypeGenerator toptypegenerator(inherited_types, top_types);
              toptypegenerator.generate();
              verbose->print("Top types:");
              verbose->print(top_types.toString());

              MultiMethodTypes db_types;
              DBTypeGenerator dbtypegenerator(inherited_types, top_types, db_types);
              dbtypegenerator.generate();
              verbose->print("DB types:");
              verbose->print(db_types.toString());

              SourceModifier *sourcemodifier;
              if (! args_info.modular_given)
                {
                  sourcemodifier = new SourceModifierDouble
                   (&sourcestructs, &modified_declarations, &modified_methods,
                    multi_method_types, db_types, inherited_types);
                }
              else
                {
                  sourcemodifier = new SourceModifierModular
                   (&sourcestructs, &modified_declarations, &modified_methods,
                    multi_method_types, db_types, inherited_types);
                }
              sourcemodifier->analyse ();

              IncludeModifier includemodifier
                (&sourcestructs, &modified_declarations);
              includemodifier.modify_prog ();

              SourceGenerator sourcegenerator
                (&sourcestructs, &modified_declarations, &modified_methods);

              sourcegenerator.generate ();

              /*
              for (SourceStructsIterator it = modified_declarations.begin ();
                          it != modified_declarations.end (); ++it)
               (*it).second->generate ();

              for (SourceStructsIterator it = modified_methods.begin ();
                          it != modified_methods.end (); ++it)
                (*it).second->generate ();
              */
            }
          else
           {
             cerr << "exit with failure" << endl;
             return res;
           }
        }
      else
        {
          cerr << "exit with failure" << endl;
          return res;
        }
    }
  else
    {
      cerr << "no input file given" << endl;
      return EXIT_FAILURE;
    }

  if (args_info.stats_given)
  {
    gettimeofday(&tv2, 0);

    double elapsed_sec = (double)(tv2.tv_sec - tv1.tv_sec) + ((double)(tv2.tv_usec - tv1.tv_usec)/1E6);

    cerr << "Analyzed files: " << num_of_analysed_files << endl;
    cerr << "Analyzed classes: " << num_of_analysed_classes << endl;
    cerr << "Elapsed time: " << elapsed_sec << endl;
  }

  return EXIT_SUCCESS;
}
