/***************************************************************************
 *   Copyright (C) 2004 by Lorenzo Bettini                                 *
 *   http://www.lorenzobettini.it                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <assert.h>

#include "inheritedtypegenerator.h"

#include "multimethodtypes.h"
#include "classdecl.h"
#include "methods.h"
#include "multimethod.h"
#include "mainerrormanager.h"
#include "methparam.h"

InheritedTypeGenerator::InheritedTypeGenerator(const MultiMethodTypes& t, MultiMethodTypes &inh)
  : MultiTypeModifier(t), inherited_types (inh), error (false)
{
}


void
InheritedTypeGenerator::generate_type(const string& meth, const MultiMethodType& t)
{
  const ClassDecl *current_receiver_class;

  for (MultiMethodTypeIterator current = t.begin(); current != t.end(); ++current)
    {
      current_receiver_class = current->getReceiverType();

      generate_type(meth, *current, current_receiver_class);
    }
}

void
InheritedTypeGenerator::generate_type(const string& method_name, const BranchType &btype, const ClassDecl *class_to_inspect)
{
  const MultiMethod *branch = class_to_inspect->retrieveMultiMethod(method_name);

  const MethParams *params = btype.getParams();
  const DeclType *spec_param = btype.getParam()->getDeclType();
  const DeclType *ret = btype.getRetType();
  const ClassDecl *starting_class = btype.getReceiverType();
  AccessType access_type = btype.getAccessType();

  if (branch)
    {
      // the branch may not exist if this class in the hierarchy
      // does not define any new method for this branch, or simply
      // because we already analysed the uppest class in the hierarchy
      // defining this branch
      const Methods *methods = branch->getMethods();

      assert (methods);

      const MethDecl *meth;
      for (MethodsIterator mit = methods->begin ();
          mit != methods->end(); ++mit)
        {
          meth = *mit;

          if (meth->equalsSignature(method_name, params, 1))
            {
              if (starting_class != class_to_inspect)
                {
                  // if we are examining a base class we have to stop
                  // if we find a private method, and we must not inspect
                  // further superclasses.
                  if (meth->getAccessType() == PRIVATE)
                    return;

                  if (!ret->equals(meth->getDeclType()))
                    {
                      const DeclType *spec_param2 = meth->getSpecializingParam()->getDeclType();

                      if (! spec_param->unrelated(spec_param2))
                        {
                          errormanager->print
                              ("return type of this branch (" + ret->getTypeName() + ")", btype);
                          errormanager->print
                              ("conflicts with return type (" + meth->getDeclType()->getTypeName() +
                              ") of branch in superclass", meth);
                          error = true;
                        }
                    }
                }

              inherited_types.addType(
                  BranchType(method_name, starting_class, meth->getSpecializingParam(),
                        meth->getDeclType(), meth->getParams(), access_type, *meth));
            }
        }
    }

  // in any case we have to inspect all the hierarchy
  const BaseClasses *bases = class_to_inspect->getBaseClasses();
  if (! bases || bases->size () < 1)
    return;

  for (BaseClassesIterator bit = bases->begin ();
      bit != bases->end (); ++bit)
    if ( (*bit)->getClassDecl () )
      generate_type (method_name, btype, (*bit)->getClassDecl ());
}
