%{
/*
 * Copyright (C) 2003 Lorenzo Bettini <http://www.lorenzobettini.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>

#include <string.h>

#include "yyerror.h"
#include "progelems.hpp"
#include "normaltext.h"
#include "mainprogelems.h"
#include "classdecl.h"
#include "multimethod.h"
#include "methodbody.h"
#include "classend.h"
#include "forwarddecls.h"
#include "virtmultimethod.h"
#include "dispatchmeths.h"
#include "includeheader.h"
#include "accesstype.h"
#include "methparams.h"

#define YYERROR_VERBOSE

int line = 0 ;
AccessType current_access_type;

extern int yylex() ;
extern FILE *yyin ;

%}

%union {
  int tok ; /* command */
  char * ide_string ; /* ide_string : id, ... */
  struct multimethodinfo {
    bool is_virtual; // whether the multimethod is virtual
    char *previous; // possible text before multimethod
  } multimethodinfo ;
  bool boolean;

  class NormalText *normaltext;
  class ProgElem *progelem;
  class ClassDecl *classdecl;
  class BaseClass *baseclass;
  BaseClasses *baseclasses;
  DerivedType derivedtype;
  class DeclType *decl_type;
  class MultiMethod *multimethod;
  class MethDecl *methdecl;
  class MethParam *methparam;
  class MethParams *methparams;
  class MethodBody *methodbody;
  MethodInstruction *methinstr;
  class ClassEnd *classend;
  class ForwardDecl *forwarddecl;
  class ForwardDecls *forwarddecls;
  class IncludeHeader *includeheader;
  class VirtMultiMethod *virtualmultimethod;
  class DispatchMeths *dispatchmeths;
  class Methods *methods;
} ;

%token <ide_string> NORMAL_T NEWLINE_T IDE_T INCLUDEFILE_T
%token <tok> PUBLIC_T PRIVATE_T PROTECTED_T VIRTUAL_T SINGLECOMMENT_T
%token <tok> END_BRANCH CONST_T BEGIN_METHOD CLASS_END INCLUDE_T
%token <tok> BEGIN_FORWARD_DECL END_TAG BEGIN_PREVIOUS_BRANCH BEGIN_DISPATCH
%token <ide_string> END_METHOD
%token <multimethodinfo> BRANCH_T
%token <boolean> CLASS_T

%type <ide_string> text
%type <boolean> isvirtual isconst ispointer
%type <progelem> possibletext
%type <normaltext> normaltext
%type <classdecl> classdecl
%type <baseclass> baseclass
%type <baseclasses> baseclasses inheritfrom
%type <derivedtype> derivedtype
%type <decl_type> decl_type
%type <multimethod> multimethod branchdecl
%type <methdecl> branchmeth method
%type <methparams> paramlist
%type <methparam> param
%type <methodbody> methbody
%type <methinstr> methinstr
%type <classend> classend
%type <forwarddecl> forwarddecl
%type <forwarddecls> forwarddecls forwardclasses
%type <includeheader> includeheader
%type <virtualmultimethod> virtualmultimethod
%type <dispatchmeths> dispatchmeths
%type <methods> methods
%type <ide_string> paramname

%%

text : possibletext
       {
               //printf("%s", $1);
        prog_elems->add_item ($1);
       }
     | text possibletext
       {
        //printf("%s", $2);
        prog_elems->add_item ($2);
       }
;

possibletext : classdecl { $$ = $1; }
             | normaltext { $$ = $1; }
             | branchdecl { $$ = $1; }
             | classend { $$ = $1; }
             | forwarddecls { $$ = $1; }
             | virtualmultimethod { $$ = $1; }
             | dispatchmeths { $$ = $1; }
             | includeheader { $$ = $1; }
             ;

classdecl : CLASS_T IDE_T inheritfrom '{'
          {
            $$ = new ClassDecl ($2, $3);
            $$->setInfo (@1.first_line);
            $$->setStruct(!$1);
          }
;

normaltext : NORMAL_T { $$ = new NormalText ($1); }
           | NEWLINE_T { $$ = new NormalText ($1); }
;

classend : CLASS_END { $$ = new ClassEnd; }
;

forwarddecls : BEGIN_FORWARD_DECL forwardclasses END_TAG
              { $$ = $2; }
;

forwardclasses : { $$ = 0; }
               | forwardclasses forwarddecl
               {
                 $1->push_back ($2);
                 $$ = $1;
               }
               | forwarddecl
               {
                 $$ = new ForwardDecls;
                 $$->push_back ($1);
               }
;

forwarddecl : CLASS_T IDE_T ';'
             {
               $$ = new ForwardDecl ($2);
             }
             |
             CLASS_T IDE_T ';' SINGLECOMMENT_T IDE_T
             {
               $$ = new ForwardDecl ($2, $5);
             }
;

includeheader : INCLUDE_T INCLUDEFILE_T
                {
                  $$ = new IncludeHeader ($2);
                }

virtualmultimethod : BEGIN_PREVIOUS_BRANCH methods END_TAG
            { $$ = new VirtMultiMethod ($2); }
;

dispatchmeths : BEGIN_DISPATCH methods END_TAG
            { $$ = new DispatchMeths ($2); }
;

inheritfrom : { $$ = 0; }
            | ':' baseclasses { $$ = $2; }
;

baseclasses : { $$ = 0; }
            | baseclasses ',' baseclass
            {
              $1->push_back ($3);
              $$ = $1;
            }
            | baseclass
            {
              $$ = new BaseClasses;
              $$->push_back ($1);
            }
;

baseclass: derivedtype isvirtual IDE_T
         {
           $$ = new BaseClass ($3, $2, $1);
           $$->setInfo (@3.first_line);
         }
         | VIRTUAL_T derivedtype IDE_T
         {
           $$ = new BaseClass ($3, true, $2);
           $$->setInfo (@3.first_line);
         }
;

isvirtual: { $$ = false; }
         | VIRTUAL_T { $$ = true; }
;

derivedtype: { $$ = NONE; }
           | PUBLIC_T { $$ = PUBLIC_TYPE; }
           | PRIVATE_T { $$ = PRIVATE_TYPE; }
           | PROTECTED_T { $$ = PROTECTED_TYPE; }
;

branchdecl: BRANCH_T IDE_T multimethod END_BRANCH
          {
                // first check whether there was text in the same lin
                // before multimethod declaration
                if ($1.previous && (strlen($1.previous) > 0)) {
                  prog_elems->add_item (new NormalText($1.previous));
                }
          
                $$ = $3;
                $$->setInfo (@1.first_line, @4.first_line - 1);
                // subtract 1 because of the \n added after the generated comment
                if ($1.is_virtual)
                  $$->setVirtual(true);
                $$->updateMethodsName ($2);
          }
;

multimethod: multimethod branchmeth
        {
          $1->add_method ($2);
          $$ = $1;
        }
           | branchmeth
        {
          $$ = new MultiMethod;
          $$->add_method ($1);
        }
;

branchmeth: isvirtual decl_type '(' paramlist ')' BEGIN_METHOD methbody
          {
            $$ = new MethDecl($2, $4, $7, current_access_type);
            $$->setInfo (@3.first_line);
            if ($1)
              $$->setVirtual(true);
            $7->setInfo (@6.first_line);
          }
          | isvirtual decl_type '(' paramlist ')' ';'
          {
            $$ = new MethDecl($2, $4, 0, current_access_type);
            $$->setInfo (@3.first_line);
            if ($1)
              $$->setVirtual(true);
          }
;

methods: methods method
        {
          if ($2) $1->push_back ($2);
          $$ = $1;
        }
        | method
        {
          $$ = new Methods;
          if ($1) $$->push_back ($1);
        }
;

method: decl_type IDE_T '(' paramlist ')' ';'
          {
            $$ = new MethDecl($1, $4, 0, current_access_type);
            $$->setMethodName ($2);
          }
      | decl_type IDE_T '(' paramlist ')' BEGIN_METHOD methbody
          {
            // simply discard it: it is a previously generated
            // _forward_ method for private and protected multimethods
            $$ = 0;
          }
;

paramlist : { $$ = 0; }
          | paramlist ',' param
            {
              $1->addParam ($3);
              $$ = $1;
            }
          | param
            {
              $$ = new MethParams;
              $$->addParam ($1);
            }
;

param: decl_type paramname
      {
        $$ = new MethParam ($2, $1);
        $$->setInfo (@1.first_line);
      }
;

paramname: { $$ = strdup (""); }
         | IDE_T { $$ = $1; }
;

decl_type: isconst IDE_T ispointer
          {
            $$ = new DeclType ($2, $1, $3);
            $$->setInfo (@2.first_line);
          }
;

isconst: { $$ = false; }
       | CONST_T { $$ = true; }
;

ispointer: { $$ = false; }
         | '*' { $$ = true; }
;

methbody: methbody methinstr
          {
            $1->addInstruction ($2);
            $$ = $1;
          }
          | methinstr
          {
            $$ = new MethodBody ();
            $$->addInstruction ($1);
          }
;

methinstr: normaltext { $$ = $1; }
         | END_METHOD { $$ = new MethodInstruction ($1); }
;

%%

