%{
/*
 * Copyright (C) 2002, Lorenzo Bettini <http://www.lorenzobettini.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <string.h>
#include <iostream>

#include "my_sstream.h"

#include "baseclass.h"
#include "methodbody.h"
#include "parser.h"
#include "tokeninfo.h"
#include "progelems.hpp"
#include "normaltext.h"
#include "mainprogelems.h"
#include "lexerfunctions.h"
#include "mainglobalvars.h"
#include "accesstype.h"
#include "mainerrormanager.h"

extern int line ;

// the current class/struct access type (PUBLIC,PROTECTED,PRIVATE)
extern AccessType current_access_type;

// whether we're inspecting a class or a struct
static bool isClass = true;

#define LINEBUF_LEN 1024

int tokenpos = 0;
char linebuf[LINEBUF_LEN] ; /* current line */

static ostringstream line_buf;

static int count_nl(const string &);

static char *flush_line_buf();

// VERB_SCAN: output information about scanning
//#define VERB_SCAN(s) s
#define VERB_SCAN(s)
#define VERB_SCAN2(s) VERB_SCAN(std::cerr << s << std::endl;)

#define PUSH(s) yy_push_state(s);
#define POP() yy_pop_state();

static int branch_bracket = 0;
// the number of brackets encountered when scanning a branch body

%}

%option stack nomain yylineno noyywrap nounput

ws [[:blank:]]+

nl \n
cr \r

IDE [a-zA-Z_]([a-zA-Z0-9_])*
BASECLASS {IDE}(<([a-zA-Z0-9_, "*"])*>)*
FILENAME ([a-zA-Z_\.]([a-zA-Z0-9_/\\])*)("."[a-zA-Z_]([a-zA-Z0-9_/\\])*)*

STRING \"[^\"\n]*\"

INCFILE (\"{FILENAME}\")|(<{FILENAME}>)

LINEINFO \#line{ws}+[0-9]+{ws}+{INCFILE}

%s CLASS_STATE BRANCH_STATE METHOD_STATE FORWARD_STATE PREVIOUS_BRANCH_STATE
%s DISPATCH_STATE CLASSDEF_STATE STRING_STATE COMMENT_STATE SINGLELINE_COMMENT
%s BRACKET_STATE INCLUDE_STATE TEMPLATE_STATE

%%

<INITIAL,CLASSDEF_STATE,BRANCH_STATE,METHOD_STATE,BRACKET_STATE>\#include{ws}* {
        PUSH (INCLUDE_STATE);
        return INCLUDE_T;
      }
<INCLUDE_STATE>{INCFILE} {
        POP();
        yylval.ide_string = strdup (yytext);
        return INCLUDEFILE_T;
      }

<INITIAL,CLASS_STATE,BRANCH_STATE,METHOD_STATE,FORWARD_STATE,PREVIOUS_BRANCH_STATE,DISPATCH_STATE,CLASSDEF_STATE,BRACKET_STATE,TEMPLATE_STATE>"/*" {
       PUSH(COMMENT_STATE) ;
       VERB_SCAN (std::cerr << "START COMMENT" << std::endl; );
       line_buf << yytext;
     }
<INITIAL,CLASS_STATE,BRANCH_STATE,METHOD_STATE,FORWARD_STATE,PREVIOUS_BRANCH_STATE,DISPATCH_STATE,CLASSDEF_STATE,BRACKET_STATE,TEMPLATE_STATE>"/*".*"*/" {
     VERB_SCAN (std::cerr << "START COMMENT" << std::endl; );
     line_buf << yytext;
     yylval.ide_string = strdup (line_buf.str().c_str());
     line_buf.str("");
     return NORMAL_T;
}
<COMMENT_STATE>"*/" {
     VERB_SCAN (std::cerr << "END COMMENT" << std::endl; );
     POP() ;
     line_buf << yytext;
     yylval.ide_string = strdup (line_buf.str().c_str());
     line_buf.str("");
     return NORMAL_T;
}

<INITIAL,CLASS_STATE,BRANCH_STATE,METHOD_STATE,CLASSDEF_STATE,BRACKET_STATE,TEMPLATE_STATE>"//" {
   PUSH (SINGLELINE_COMMENT) ;
   VERB_SCAN (std::cerr << "START COMMENT" << std::endl; );
   line_buf << yytext;
}
<SINGLELINE_COMMENT>{nl} {
   VERB_SCAN (std::cerr << "END COMMENT" << std::endl; );
   POP() ;
   ++line;
   line_buf << yytext;
   yylval.ide_string = strdup (line_buf.str().c_str());
   line_buf.str("");
   return NORMAL_T;
}

<INITIAL>"template"({ws}|{nl})*"<"(\n|.)*">" {
    yylval.ide_string = strdup (yytext);
    line += count_nl(yytext);
    return NORMAL_T;
}

<INITIAL>"class"{ws}+{IDE}{ws}*";" {
        VERB_SCAN (std::cerr << "FORWARD DECL: " << yytext << std::endl; )
        line_buf << yytext;
}

<INITIAL>"struct"{ws}+{IDE}{ws}*";" {
        VERB_SCAN (std::cerr << "FORWARD DECL: " << yytext << std::endl; )
        line_buf << yytext;
}

<INITIAL>"class " {
        PUSH(CLASS_STATE);
        VERB_SCAN( std::cerr << "CLASS" << std::endl; )
        updateTokenInfo();
        isClass = true;
        yylval.boolean = true; // it's a class and not a struct
        return CLASS_T;
  }
  
<INITIAL>"struct " {
        PUSH(CLASS_STATE);
        VERB_SCAN( std::cerr << "CLASS" << std::endl; )
        updateTokenInfo();
        isClass = false;
        yylval.boolean = false; // it's a struct and not a class
        return CLASS_T;
  }
  
<CLASS_STATE>"public" { return PUBLIC_T; }
<CLASS_STATE>"private" { return PRIVATE_T; }
<CLASS_STATE>"protected" { return PROTECTED_T; }
<CLASS_STATE,BRANCH_STATE>"virtual" { return VIRTUAL_T; }
<CLASS_STATE>":" { VERB_SCAN( std::cerr << yytext << std::endl; ) return ':'; }
<CLASS_STATE>"," { VERB_SCAN( std::cerr << yytext << std::endl; ) return ','; }
<CLASS_STATE>{BASECLASS} {
        updateTokenInfo ();
        yylval.ide_string = strdup (yytext);
        VERB_SCAN( std::cerr << "BASECLASS: " << yytext << std::endl; )
        return IDE_T;
  }
<CLASS_STATE>{ws} {
        // std::cerr << "SPACE" << std::endl;
  }
<CLASS_STATE,BRANCH_STATE,PREVIOUS_BRANCH_STATE,FORWARD_STATE,DISPATCH_STATE>{nl} { ++line; }
<CLASS_STATE>"{" { 
  PUSH(CLASSDEF_STATE); 
  current_access_type = (isClass ? PRIVATE : PUBLIC); 
  return '{'; 
}

<CLASSDEF_STATE>"public:" {
        current_access_type = PUBLIC;
        VERB_SCAN2("PUBLIC")
        line_buf << yytext;
        yylval.ide_string = strdup (line_buf.str().c_str());
        line_buf.str("");
        return NORMAL_T;
}
<CLASSDEF_STATE>"protected:" {
        current_access_type = PROTECTED;
        VERB_SCAN2("PROTECTED")
        line_buf << yytext;
        yylval.ide_string = strdup (line_buf.str().c_str());
        line_buf.str("");
        return NORMAL_T;
}
<CLASSDEF_STATE>"private:" {
        current_access_type = PRIVATE;
        VERB_SCAN2("PRIVATE")
        line_buf << yytext;
        yylval.ide_string = strdup (line_buf.str().c_str());
        line_buf.str("");
        return NORMAL_T;
}

<CLASSDEF_STATE>{ws}*"virtual"({ws}|{nl})*"branches " {
        PUSH(BRANCH_STATE);
        VERB_SCAN( std::cerr << "VIRTUAL BRANCH" << std::endl; )
        line += count_nl(yytext);
        updateTokenInfo ();
        yylval.multimethodinfo.previous = flush_line_buf();
        yylval.multimethodinfo.is_virtual = true;
        return BRANCH_T;
}
<CLASSDEF_STATE>{ws}*"virtual"({ws}|{nl})*"branch " {
        PUSH(BRANCH_STATE);
        VERB_SCAN( std::cerr << "VIRTUAL BRANCH" << std::endl; )
        line += count_nl(yytext);
        updateTokenInfo ();
        yylval.multimethodinfo.previous = flush_line_buf();
        yylval.multimethodinfo.is_virtual = true;
        return BRANCH_T;
}

<CLASSDEF_STATE>{ws}*"branches " {
        PUSH(BRANCH_STATE);
        VERB_SCAN( std::cerr << "BRANCH" << std::endl; )
        updateTokenInfo ();
        yylval.multimethodinfo.previous = flush_line_buf();
        yylval.multimethodinfo.is_virtual = false;
        return BRANCH_T;
}
<CLASSDEF_STATE>{ws}*"branch " {
        PUSH(BRANCH_STATE);
        VERB_SCAN( std::cerr << "BRANCH" << std::endl; )
        updateTokenInfo ();
        yylval.multimethodinfo.previous = flush_line_buf();
        yylval.multimethodinfo.is_virtual = false;
        return BRANCH_T;
}

<CLASSDEF_STATE,BRACKET_STATE,METHOD_STATE>\" {
        PUSH(STRING_STATE) ;
        VERB_SCAN (std::cerr << "STRING" << std::endl; );
        line_buf << yytext;
}
<CLASSDEF_STATE,BRACKET_STATE>"{" {
        PUSH (BRACKET_STATE) ;
        VERB_SCAN (std::cerr << "BRACKET" << std::endl; );
        line_buf << yytext;
}
<BRACKET_STATE>"}" {
        POP ();
        VERB_SCAN (std::cerr << "END BRACKET" << std::endl; );
        line_buf << yytext;
}
<STRING_STATE>\" {
        POP() ;
        VERB_SCAN (std::cerr << "END STRING" << std::endl; );
        line_buf << yytext;
}
<BRANCH_STATE>"endbranches" {
        POP();
        updateTokenInfo();
        VERB_SCAN( std::cerr << "END BRANCH" << std::endl; )
        return END_BRANCH;
}
<BRANCH_STATE>"endbranch" {
        POP();
        updateTokenInfo();
        VERB_SCAN( std::cerr << "END BRANCH" << std::endl; )
        return END_BRANCH;
}
<BRANCH_STATE,PREVIOUS_BRANCH_STATE,DISPATCH_STATE>"const" { VERB_SCAN( std::cerr << "CONST" << std::endl; ) return CONST_T; }
<BRANCH_STATE,PREVIOUS_BRANCH_STATE,DISPATCH_STATE>"(" { updateTokenInfo (); return '('; }
<BRANCH_STATE,PREVIOUS_BRANCH_STATE,DISPATCH_STATE>")" { VERB_SCAN( std::cerr << ")" << std::endl; ) return ')'; }
<BRANCH_STATE,PREVIOUS_BRANCH_STATE,DISPATCH_STATE>"," { return ','; }
<BRANCH_STATE,PREVIOUS_BRANCH_STATE>"{" { PUSH (METHOD_STATE); updateTokenInfo (); return BEGIN_METHOD; }
<BRANCH_STATE>";" { VERB_SCAN(std::cerr << "END METH DECL" << std::endl); return ';'; }
<BRANCH_STATE,PREVIOUS_BRANCH_STATE,DISPATCH_STATE>"*" { VERB_SCAN( std::cerr << "POINTER" << std::endl; ) return '*'; }
<BRANCH_STATE,PREVIOUS_BRANCH_STATE,DISPATCH_STATE,FORWARD_STATE>{ws} {
        // std::cerr << "SPACE" << std::endl;
}
<BRANCH_STATE>{IDE} {
        updateTokenInfo ();
        yylval.ide_string = strdup (yytext);
        VERB_SCAN( std::cerr << "IDE: " << yytext << ", line: " << yylloc.first_line << std::endl; )
        return IDE_T;
}

<METHOD_STATE>"}"({ws}|{nl})*";" {
        POP();
        // return also the possible last buffered instruction, i.e.,
        // when "};" is not on a new line.  E.g., "{ foo(); };",
        // " foo(); " is still in the line_buf

        yylval.ide_string = strdup (line_buf.str().c_str());
        line_buf.str("");
        line += count_nl(yytext);
        return END_METHOD;
  }
<METHOD_STATE>"{" {
    ++branch_bracket;
    line_buf << yytext;
  }
<METHOD_STATE>"}" {
    if (branch_bracket == 0)
    {
      // ERROR: we miss the terminating ;
      errormanager->print
          ("incorrect branch ending", input_file, line);
      errormanager->exit_failure();
    }

    --branch_bracket;
    line_buf << yytext;
  }
<METHOD_STATE>{nl} {
        ++line;
        // std::cerr << "NL" << std::endl;
        // std::cerr << "line: " << line_buf.str() << std::endl;
        line_buf << yytext;
        yylval.ide_string = strdup (line_buf.str().c_str());
        line_buf.str("");
        return NORMAL_T;
  }
<METHOD_STATE>. {
        //std::cerr << "METHOD NORMAL TEXT: " << yytext << std::endl;
        line_buf << yytext;
  }

<INITIAL>"// doublecpp: forward declarations".*{nl} {
          ++line;
          PUSH(FORWARD_STATE);
          VERB_SCAN( std::cerr << "FORWARD_STATE" << std::endl; )
          return BEGIN_FORWARD_DECL;
        }
<FORWARD_STATE>"// doublecpp: end".*{nl}{nl} {
          line += 2;
          POP();
          VERB_SCAN(std::cerr << "END FORWARD_STATE" << std::endl;)
          return END_TAG;
        }
<FORWARD_STATE>"// file:" {
          // std::cerr << "// file:" << std::endl;
          return SINGLECOMMENT_T;
        }
<FORWARD_STATE>"class " {
          /* for forwarddecls. generated by doublecpp, there's actually no
             need to specify whether it's a class or a struct, since they'll
             be used only by sourceanalyser to inspect other files. */
          VERB_SCAN( std::cerr << "CLASS FORWARD_STATE" << std::endl; )
          yylval.boolean = true; // it's a class
          return CLASS_T;
        }
<FORWARD_STATE>"struct " {
          VERB_SCAN( std::cerr << "STRUCT FORWARD_STATE" << std::endl; )
          yylval.boolean = false; // it's a class
          return CLASS_T;
        }
<FORWARD_STATE>{FILENAME} {
        yylval.ide_string = strdup (yytext);
        // std::cerr << "FILENAME: " << yytext << std::endl;
        return IDE_T;
}

<CLASSDEF_STATE>"// doublecpp: method branches".*{nl} {
          ++line;
          PUSH(PREVIOUS_BRANCH_STATE);
          VERB_SCAN(std::cerr << "PREVIOUS_BRANCH_STATE" << std::endl;)
          return BEGIN_PREVIOUS_BRANCH;
        }
<PREVIOUS_BRANCH_STATE>"public:" {
        // simply skip it during scanning previously generated files
        VERB_SCAN(std::cerr << "public:" << std::endl;)
}
<PREVIOUS_BRANCH_STATE>"private:" {
        VERB_SCAN(std::cerr << "private:" << std::endl;)
        }
<PREVIOUS_BRANCH_STATE>"protected:" {
        VERB_SCAN(std::cerr << "protected:" << std::endl;)
        }
<PREVIOUS_BRANCH_STATE>{LINEINFO} {
        VERB_SCAN(std::cerr << "line info " << yytext << std::endl;)
        }
<PREVIOUS_BRANCH_STATE,DISPATCH_STATE,FORWARD_STATE>";" {
          // std::cerr << ";" << std::endl;
          return ';';
        }
<PREVIOUS_BRANCH_STATE,DISPATCH_STATE>"virtual" {
          //std::cerr << "virtual" << std::endl;
        }
<PREVIOUS_BRANCH_STATE>"using "{IDE}.*{nl} {
          ++line;
          VERB_SCAN(std::cerr << "discarding using" << std::endl;)
        }
<PREVIOUS_BRANCH_STATE,DISPATCH_STATE>{IDE} {
        yylval.ide_string = strdup (yytext);
        // std::cerr << "IDE: " << yytext << std::endl;
        return IDE_T;
}
<PREVIOUS_BRANCH_STATE,DISPATCH_STATE>"// doublecpp: end".*{nl} {
          ++line;
          POP();
          VERB_SCAN(std::cerr << "END" << std::endl;)
          return END_TAG;
        }

<CLASSDEF_STATE>"// doublecpp: dispatch methods".*{nl} {
          ++line;
          PUSH(DISPATCH_STATE);
          VERB_SCAN(std::cerr << "DISPATCH_STATE" << std::endl;)
          return BEGIN_DISPATCH;
        }
<DISPATCH_STATE>"public:".*{nl} {
          ++line;
          //std::cerr << "discarding public" << std::endl;
        }

<CLASSDEF_STATE>"}"{ws}*";" {
        BEGIN INITIAL;
        VERB_SCAN( std::cerr << "CLASS_END" << std::endl; )
        return CLASS_END;
}

<CLASSDEF_STATE>"}" {
        errormanager->print
          ("incorrect class ending", input_file, line);
        errormanager->exit_failure();
        return CLASS_END;
}

<INITIAL,CLASSDEF_STATE,STRING_STATE,COMMENT_STATE,BRACKET_STATE>{nl} {
        ++line;
        // std::cerr << "NL" << std::endl;
        VERB_SCAN( std::cerr << "line: " << line_buf.str() << std::endl; )
        line_buf << yytext;
        yylval.ide_string = strdup (line_buf.str().c_str());
        line_buf.str("");
        return NORMAL_T;
  }

<<EOF>> {
        VERB_SCAN( std::cerr << "EOF" << std::endl; )
        // std::cerr << "line: " << line_buf.str() << std::endl;
        BEGIN INITIAL;
        prog_elems->add_item(new NormalText (line_buf.str()));
        line_buf.str("");
        yyterminate();
  }

<INITIAL,CLASSDEF_STATE,STRING_STATE,COMMENT_STATE,SINGLELINE_COMMENT,BRACKET_STATE>. {
        //std::cerr << "NORMAL TEXT: " << yytext << std::endl;
        line_buf << yytext;
  }


%%

void
initialize_lexer(const string &file)
{
  input_file = file;
  line = 1;
}

int
count_nl(const string &s)
{
  int count = 0;
  string app = s;
  string::size_type pos;

  while ((pos = app.find("\n")) != string::npos)
  {
    ++count;
    app = app.substr(pos+1);
  }

  return count;
}

/**
  flushes the line buffer and returns the content of the buffer
*/
char *
flush_line_buf()
{
  char * buff = strdup (line_buf.str().c_str());
  line_buf.str("");
  return buff;
}
