/***************************************************************************
 *   Copyright (C) 2003 by Lorenzo Bettini                                 *
 *   http://www.lorenzobettini.it                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <cstdlib>
#include "sourceanalyser.h"
#include "fileutil.h"
#include "progelems.hpp"
#include "mainprogelems.h"
#include "mainglobalvars.h"
#include "mainverbose.h"
#include "classdecls.h"
#include "mainerrormanager.h"
#include "lexerfunctions.h"
#include "classdecl.h"
#include "methparam.h"

extern int yyparse();

using namespace std;

static const string
get_file_name(const string &);

static bool
can_convert_to_file_name(const string &);

// if the class contains <>, e.g. for template classes, set<string>
bool
can_convert_to_file_name(const string &name)
{
  return (name.find ("<") == string::npos);
}

int
SourceAnalyser::analyse_program(const string &file_name)
{
    if (analyse (file_name) == EXIT_FAILURE)
        return EXIT_FAILURE;

    return analyse_previous_outputs ();
}

int
SourceAnalyser::analyse(const string &file_name, bool previous_output)
{
  if (analysed_files.find (file_name) != analysed_files.end ())
    return EXIT_SUCCESS; // already analysed

  analysed_files.insert (file_name);
  ++scanned_files;

  verbose->print
    ("inspecting file: " + file_name);
  verbose->inc_indent();

  input_file =
    (contains_path (file_name) ? string (file_name) : main_path + file_name);
  open_input_file (input_file);

  const string stripped_file_name =
    (contains_path (input_file) ? strip_file_path (input_file) : input_file);
  const string file_name_no_ext =
    get_file_name_no_ext (stripped_file_name);

  prog_elems = new ProgElems
   (file_name_no_ext, stripped_file_name, cout);

  initialize_lexer (file_name);
  if (yyparse () == EXIT_FAILURE)
    return EXIT_FAILURE;

  ProgElems *current_source = prog_elems;
  // since this will be called recursively from now on,
  // we have to save the current_source

  // first visit the header files
  const IncludeHeaders &headers = current_source->getIncludeHeaders ();
  if (headers.size () > 0)
    {
      for (IncludeHeaderIterator hit = headers.begin();
           hit != headers.end(); ++hit)
        if (analyse_header (*hit) == EXIT_FAILURE)
            return EXIT_FAILURE;
        }

  const ClassDecls &classes = current_source->getClasses ();
  /*
  if (classes.size () == 0)
    {
      errormanager->print
        ("no class defined in this source");
      return EXIT_FAILURE;
    }
    */

  if (classes.size ())
    {
      for (ClassDeclIterator cit = classes.begin ();
           cit != classes.end (); ++cit)
        {
          if (analyse_class (*cit, current_source) == EXIT_FAILURE)
            return EXIT_FAILURE;
        }
    }
  else
    sources->addProgElem(current_source);

  // now inspect a possible previously generated output
  // in order not to overwrite parts that were not generated
  // for this hierarchy

  if (previous_output)
    if (analyse_previous_output (current_source) == EXIT_FAILURE)
      return EXIT_FAILURE;

  verbose->dec_indent();
  return EXIT_SUCCESS;
}

int
SourceAnalyser::analyse_previous_outputs()
{
  verbose->print
    ("inspecting previously generated files...");

  const ProgElemsList &source_list = sources->getProgElemsList ();
  for (ProgElemsListIterator it = source_list.begin();
       it != source_list.end(); ++it)
    if (analyse_previous_output (*it) == EXIT_FAILURE)
        return EXIT_FAILURE;

  return EXIT_SUCCESS;
}

int
SourceAnalyser::analyse_previous_output(const ProgElems *source)
{
  bool has_to_invade_target =
    invade_target && ! source->contains_multimethods();
  const string previous_output =
        main_path + source->getName () +
        ( has_to_invade_target ? "" : output_header_file_suff) +
        "." +
        ( has_to_invade_target ?
          input_header_file_ext : output_header_file_ext);

  if (open_input_file_if_exists (previous_output))
    {
       ++scanned_files;

       verbose->print
        ("inspecting previously generated file: " + previous_output);
       prog_elems = new ProgElems("foo", "foo", cout);
       ProgElems *temp_prog_elems = prog_elems;
       initialize_lexer (previous_output);
       if (yyparse () == EXIT_FAILURE)
         return EXIT_FAILURE;

       const ForwardDecls &forwarddecls =
         prog_elems->getForwardDecls ();

       const ForwardDecl *fdecl = 0;
       for (ForwardDeclsIterator fit = forwarddecls.begin ();
            fit != forwarddecls.end (); ++fit)
         {
           fdecl = *fit;
           const string &filename = fdecl->getFileName();
           if (filename != "")
             analyse (main_path + filename, true);
         }

       /*
       instead of visiting the (previously) generated methods
       we visit the forward class declarations.  This should do.

       const GeneratedMethDeclList *meth_decls =
         prog_elems->getGeneratedMethDeclList();

       for (GeneratedMethDeclListIt it = meth_decls->begin ();
            it != meth_decls->end (); ++it)
         if (analyse_methods ((*it)->getMethods()) == EXIT_FAILURE)
           return EXIT_FAILURE;
       */

       delete temp_prog_elems;
     }

  return EXIT_SUCCESS;
}

int
SourceAnalyser::analyse_header(const IncludeHeader *hi)
{
  if (hi->isSystemInclude())
    return EXIT_SUCCESS;

  const string &header_file_name = hi->getFileName();

  if (excludeheaders.find(header_file_name) != excludeheaders.end())
    {
      verbose->print("skipping header file " + header_file_name +
                     " as explicitly requested");
      return EXIT_SUCCESS;
    }

  string file_name = main_path + header_file_name;

  if (analyse_header_file(file_name) == EXIT_SUCCESS)
    return EXIT_SUCCESS;
  else
    {
      for (sourcepath_iterator sit = sourcepath.begin();
           sit != sourcepath.end(); ++sit)
        if (analyse_header_file(*sit + "/" + header_file_name) == EXIT_SUCCESS)
          return EXIT_SUCCESS;
    }

  errormanager->print
      ("can't find header file " + header_file_name, hi);

  return EXIT_FAILURE;
}

int
SourceAnalyser::analyse_header_file(const string &file_name)
{
  const string header_winput_ext =
    strip_suffix (get_file_name_no_ext (file_name),
                  output_header_file_suff) + "." +
    input_header_file_ext;

  // first try with the header file, with the same
  // extension of input files
  string file_to_open = file_name;
  if (exist_file (header_winput_ext))
    file_to_open = header_winput_ext;
  else if (! exist_file(file_name))
    return EXIT_FAILURE;

  verbose->print
    ("inspecting header file: " + file_to_open);

  if (analyse (file_to_open) == EXIT_FAILURE)
    return EXIT_FAILURE;

  return EXIT_SUCCESS;
}

int
SourceAnalyser::analyse_class(const ClassDecl *classdecl, ProgElems *source)
{
   const string &class_name = classdecl->getClassName ();

   if (analysed_classes.find (class_name) !=
       analysed_classes.end ())
     return EXIT_SUCCESS; // already analysed

   analysed_classes.insert (class_name);
   ++scanned_classes;

   sources->addProgElem (class_name, source);

   if (analyse_base_classes (classdecl) == EXIT_FAILURE)
     return EXIT_FAILURE;

   const MultiMethods *multimethods = classdecl->getMultiMethods ();
   if (multimethods)
     if (analyse_multimethods (multimethods) == EXIT_FAILURE)
       return EXIT_FAILURE;

   return EXIT_SUCCESS;
}

int
SourceAnalyser::analyse_base_classes(const ClassDecl *cl)
{
  const BaseClasses *base_classes = cl->getBaseClasses ();;

  if (! base_classes)
    return EXIT_SUCCESS;

  BaseClass *base_class;
  ClassDecl *cl_decl;

  for (BaseClassesIterator bit = base_classes->begin ();
       bit != base_classes->end (); ++bit)
    {
      base_class = *bit;
      const string &base_name = base_class->getClassName ();

      if (base_class->is_template_base_class ())
        {
          errormanager->warning
           ("skipping template super class " + base_name, base_class);
          continue;
        }

      cl_decl = sources->retrieve_class_decl (base_name);

      if (! cl_decl)
        {
          errormanager->print
           ("can't find base class " + base_name,
            base_class);
          errormanager->exit_failure ();
        }

      base_class->setClassDecl (cl_decl);
    }

  return EXIT_SUCCESS;
}

int
SourceAnalyser::analyse_type_name(const string &name, const ProgElem *elem)
{
  if (analysed_classes.find (name) != analysed_classes.end ())
    return EXIT_SUCCESS; // already analysed

  // simply skip these classes, e.g., set<string>
  if (! can_convert_to_file_name (name))
    return EXIT_SUCCESS;

  if (dont_infer_headers)
    {
      // we hope to find it later
      return EXIT_SUCCESS;

      // otherwise an error will be raised during parameter binding
    }

  const string &file_name = get_file_name (name);

  if (! exist_file (file_name))
    {
      errormanager->print
       ("can't find input file " + file_name +
        " for class " + name, elem);
      errormanager->exit_failure ();
    }

  return analyse (file_name);
}

int
SourceAnalyser::analyse_method(const MethDecl *meth)
{
  // only the first parameter is relevant to double dispatch
  // TODO: check the other parameters?
  MethParam *param = meth->getSpecializingParam();

  if (! param)
    {
      errormanager->print
        ("missing parameter",
        meth);
      errormanager->exit_failure ();
    }

  const DeclType *decl_type = param->getDeclType();
  const string &param_type = decl_type->getTypeName ();
  if (analyse_type_name (param_type, decl_type) == EXIT_FAILURE)
    return EXIT_FAILURE;

  return EXIT_SUCCESS;
}

void
SourceAnalyser::generate()
{
  for (SourceStructsIterator it = sources->begin ();
       it != sources->end (); ++it)
    (*it).second->generate ();
}

const string
get_file_name(const string &name)
{
  return main_path + to_lower (name) + "." + input_header_file_ext;
}
