/***************************************************************************
 *   Copyright (C) 2003 by Lorenzo Bettini                                 *
 *   http://www.lorenzobettini.it                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <assert.h>
#include <cstdlib>
#include <iostream> // TODO: remove it when changed ProgElems cons

#include "my_sstream.h"

#include "sourcemodifier.hpp"
#include "sourcestructs.h"
#include "classend.h"
#include "mainglobalvars.h"
#include "fileutil.h"
#include "methods.h"
#include "multimethods.h"
#include "includeheader.h"
#include "mainerrormanager.h"
#include "multimethodtypes.h"
#include "methodbody.h"
#include "methparams.h"

using std::ostringstream;

SourceModifier::SourceModifier(SourceStructs *s, SourceStructs *src,
                               SourceStructs *m,
                               const MultiMethodTypes &mtypes,
                               const MultiMethodTypes &db_t,
                               const MultiMethodTypes &ext) :
  sourcestructs (s), modified_declarations (src),
  modified_methods (m),
  multi_method_types (mtypes),
  db_types (db_t),
  extended_types (ext),
  visit_only_targets (true)
{
}

SourceModifier::~SourceModifier()
{
}

#include "mainverbose.h"

int
SourceModifier::analyse()
{
  ProgElems *prog_elems;
  const ProgElemsList &progelemslist = sourcestructs->getProgElemsList ();

  for (ProgElemsListIterator sit = progelemslist.begin ();
       sit != progelemslist.end (); ++sit)
    {
      prog_elems = *sit;

      if (prog_elems->contains_multimethods())
        analyse_source (prog_elems);
    }

  if (! visit_only_targets && ! clean_targets)
    return EXIT_SUCCESS;

  // the sources that contain only targets are modified later,
  // when all the other classes are already modified.  This is useful
  // in case of invade_target options, since we will insert dispatch
  // method definitions in the specializing classes modified sources
  for (ProgElemsListIterator sit = progelemslist.begin ();
       sit != progelemslist.end (); ++sit)
    {
      prog_elems = *sit;
      if (! prog_elems->contains_multimethods())
        analyse_source (prog_elems);
    }

  return EXIT_SUCCESS;
}

void
SourceModifier::analyse_source(ProgElems *source)
{
  bool to_generate = should_modify (source);

  const ClassDecls &classes = source->getClasses ();

  bool has_no_multimethods = ! source->contains_multimethods();
  ProgElems *modified_decl = createModifiedDeclSource (source, has_no_multimethods);
  ProgElems *modified_meth = createModifiedMethSource (source, has_no_multimethods);

  modified_decl->setToGenerate(to_generate);

  if (! (invade_target && has_no_multimethods))
   {
     generate_include (modified_decl->getName (), modified_meth,
                    ! (invade_target && has_no_multimethods));
     modified_meth->setToGenerate (to_generate);
   }
  // otherwise that source must remain empty

  ClassDecl *cl;
  if (classes.size ())
    {
      for (ClassDeclIterator cit = classes.begin ();
           cit != classes.end (); ++cit)
        {
          cl = *cit;
          string class_name = cl->getClassName ();
          modified_declarations->addProgElem (class_name, modified_decl);
          modified_methods->addProgElem (class_name, modified_meth);
          modified_decl->add_class_decl (cl);
          modified_meth->add_class_decl (cl);
        }
    }
  else
    modified_declarations->addProgElem (modified_decl);

  ProgElem *elem;

  for (ProgElemsIterator pit = source->begin ();
       pit != source->end (); ++pit)
    {
      elem = *pit;
      if (elem)
        transform_elem_DB (elem, modified_decl, modified_meth);
    }
}

void
SourceModifier::transform_elem(ProgElem *elem, ProgElems *modified_decl, ProgElems *modified_meth)
{
      if (! elem)
        return;

      modified_decl->add_item (new ProgElem (*elem));
      // simply add a copy of the prog elem
}

void
SourceModifier::transform_elem(MultiMethod *br, ProgElems *modified_decl, ProgElems *modified_meth)
{
      transform_multimethod (br, modified_decl, modified_meth);
}

void
SourceModifier::transform_elem(ClassDecl *cl, ProgElems *modified_decl, ProgElems *modified_meth)
{
      transform_class_decl (cl, modified_decl);
}

void
SourceModifier::transform_elem(ClassEnd *cend, ProgElems *modified_decl, ProgElems *modified_meth)
{
      transform_classend (cend, modified_decl, modified_meth);
}

void
SourceModifier::transform_elem(IncludeHeader *ih, ProgElems *modified_decl, ProgElems *modified_meth)
{
      transform_includeheader (ih, modified_decl);
}

bool
SourceModifier::should_invade_target(const string &class_name)
{
  return (invade_target && ! sourcestructs->contains_multimethods(class_name));
}

bool
SourceModifier::should_use_output_extension(const string &class_name)
{
  return (! should_invade_target(class_name));
}

// if the source has not to be invaded, then it will surely have
// to be modified (either invade_target is not specified or
// the source does not contain only targets)
bool
SourceModifier::should_modify(ProgElems *source)
{
  if (! (invade_target && ! source->contains_multimethods()))
    return true;

  // actually checks that some target is contained in this source
  const ClassDecls &classes = source->getClasses ();
  const ClassDecl *cl;
  for (ClassDeclIterator cit = classes.begin ();
       cit != classes.end (); ++cit)
    {
      cl = *cit;
      if (extended_types.contains_Param(cl))
        return true;
    }

  return false;
}

ProgElems *
SourceModifier::createModifiedDeclSource(ProgElems *source, bool has_no_multimethods)
{
  const string &source_name = source->getName();
  if (invade_target && has_no_multimethods)
    return new ProgElems
      (source_name, source_name + "." + input_header_file_ext, std::cout);
  else
    return new ProgElems
      (source_name, source_name +
       output_header_file_suff + "." + output_header_file_ext, std::cout);
}

ProgElems *
SourceModifier::createModifiedMethSource(ProgElems *source, bool has_no_multimethods)
{
  const string &source_name = source->getName();
  return new ProgElems
    (source_name, source_name +
     output_source_file_suff + "." + output_source_file_ext, std::cout);
}

void
SourceModifier::generate_include(const string &cl, ProgElems *source, bool use_output_ext)
{
  string header_file = to_lower (cl) +
    (use_output_ext ? output_header_file_suff : "" ) + "." +
    (use_output_ext ? output_header_file_ext : input_header_file_ext);

  if (source->contains_header (header_file))
    return;

  IncludeHeader *ih = new IncludeHeader("\"" + header_file + "\"");

  source->add_item_front (new NormalText ("\n\n"));
  source->add_item_front (ih);
}

void
SourceModifier::generate_include_header(const string &cl, ProgElems *source,
                                        bool use_output_ext)
{
  ProgElems *orig_source = sourcestructs->retrieve_source (cl);

  assert (orig_source);

  generate_include (orig_source->getName (), source, use_output_ext);
}

void
SourceModifier::transform_multimethod(MultiMethod *mm,
    ProgElems *modified_decl, ProgElems * modified_meth)
{
  modified_decl->add_item
   (new NormalText
    (generate_comment_tag (METHOD_BRANCHES)));

  const ClassDecl *cl = mm->getClassDecl ();

  assert (cl);

  const string class_name = cl->getClassName ();

  generate_using (cl, mm);
  generate_usings (modified_decl, mm);

  const Methods *methods = mm->getMethods ();
  MethDecl *method;
  const string &method_name = mm->getMethodName ();

  bool exist_restricted_method = false;

  if (methods && methods->size()) // TODO: should always contain something, check it before
    {
      for (MethodsIterator mit = methods->begin ();
           mit != methods->end (); ++mit)
        {
          // this is the modification for the class declaration
          method = *mit;

          if (method->isRestricted())
            exist_restricted_method = true;

          const string decl_type_string = method->getDeclType()->toString();

          const MethParams *params = method->getParams ();
          string params_string;
          params_string = params->toString ();

          const MethParam *specialized_arg_type = (*(params->begin()));
          const string specialized_arg =
            specialized_arg_type->getDeclType()->getTypeName();

          insert_line_info (modified_decl, cl->getClassName(), method);

          ostringstream meth_decl;

          if (method->isVirtual())
            meth_decl << "virtual ";

          meth_decl << decl_type_string << " ";
          meth_decl << change_method_name (method) << "(";

          meth_decl << params_string;

          modified_decl->add_item
            (new NormalText
              (meth_decl.str () + ");\n"));

          const MethodBody *body = method->getBody ();

          if (body) // otherwise it is a simple declaration
            {
              generate_include_header
                (specialized_arg, modified_meth,
                  should_use_output_extension(specialized_arg));
              modified_meth->add_item
                (new NormalText (decl_type_string + "\n"));
              insert_line_info (modified_meth, class_name, method);
              modified_meth->add_item
                (new NormalText (class_name + "::" +
                                  change_method_name (method) +
                                  "(" + params_string +")\n"));

              const MethodInstructions *instructions =
                body->getInstructions ();

              insert_line_info (modified_meth, class_name, body);
              modified_meth->add_item (new NormalText ("{"));

              for (InstructionIterator bit = instructions->begin ();
                    bit != instructions->end (); ++bit)
                modified_meth->add_item (new ProgElem (**bit));

              modified_meth->add_item (new NormalText ("}\n\n"));
            }
        }
    }

  verbose->print(method_name + "_DB methods in class " + class_name);
  verbose->print(db_types.toString(method_name, cl));

  // now deal with possibly restricted methods
  if (exist_restricted_method)
    {
      handle_restricted_methods (modified_decl, method_name, cl);
    }

  for (MultiMethodTypeIterator bit = db_types.begin(method_name, cl);
        bit != db_types.end(method_name, cl); ++bit)
    add_main_method(mm, *bit, modified_decl, modified_meth);

  insert_line_info (modified_decl, cl->getClassName(), mm, true);

  modified_decl->add_item
   (new NormalText
    (generate_comment_tag (END_TAG)));
}

void
SourceModifier::add_main_method(const MultiMethod *multimethod,
                                const BranchType &btype,
                                ProgElems *modified_decl,
                                ProgElems *modified_meth)
{
  bool is_virtual = (assume_virtual || multimethod->isVirtual());

  const string method_name = multimethod->getMethodName ();
  const string decl_type_string = btype.getRetType()->toString();

  const MethParams *params = btype.getParams();

  const MethParam *specialized_arg_type = btype.getParam();
  string top_spec_arg_param_string =
       specialized_arg_type->toString ();

  if (params->size () > 1)
    top_spec_arg_param_string += ", " + params->toString (2);

  modified_decl->add_item
    (new NormalText
    ((is_virtual ? "virtual " : "") +
      decl_type_string + " " +
      change_main_method_name (method_name)
      + "(" + top_spec_arg_param_string +
      ");\n"));

  generate_main_method(modified_meth, btype);
}

void
SourceModifier::fill_main_method_generator(main_method_gen_class &main_meth_gen,
                                           const BranchType &bt)
{
  main_meth_gen.set_ret_type (bt.getRetType()->toString());
  main_meth_gen.set_class_name (bt.getReceiverType()->getClassName());
  main_meth_gen.set_meth_name
    (change_main_method_name (bt.getMethodName()));
  main_meth_gen.set_param_string (bt.getParams()->toString());
}

// return the base classes that introduced or specialized a multimethod
void
SourceModifier::generate_using(const ClassDecl *cl, MultiMethod *branch) const
{
  const BaseClasses *bases = cl->getBaseClasses ();
  const MultiMethod *base_branch;
  AccessType accesstype = branch->getAccessType();
  const string &branch_name = branch->getMethodName();
  const ClassDecl *branch_class = branch->getClassDecl();
  // in the first call it will be the same as cl

  if (bases)
    {
      for (BaseClassesIterator bit = bases->begin ();
           bit != bases->end (); ++bit)
        {
           const BaseClass *base = *bit;

           if (base->is_template_base_class ())
             continue; // TODO: handle template base classes

           const ClassDecl *base_class = base->getClassDecl ();
           base_branch = base_class->retrieveMultiMethod(branch_name);
           bool keep_on_searching = true;

           if (base_branch)
             {
               // we have to stop in case the base classes where the multi method
               // is defined with a more restricted access type, and it's not protected

               AccessType base_access_type = base_branch->getAccessType();
               if (! stronger_access_type (base_access_type, accesstype))
                 {
                   branch->addInheritFromBase(base_class->getClassName());
                 }
               else
                 {
                   if (base_access_type == PROTECTED)
                     {
                       // if it's protected it will be inherited in the derived class
                       branch->addInheritFromBase(base_class->getClassName());
                     }
                   else
                     {
                       errormanager->warning
                         ("method " + branch_name + " is " +
                          translate_access_type (base_access_type) +
                          " in " + base_class->getClassName (), base_branch);
                       errormanager->warning
                         ("thus it won't be visible in " +
                          branch_class->getClassName(), branch);
                     }
                 }

               keep_on_searching = false;
             }

           if (keep_on_searching)
             {
               generate_using (base_class, branch);
             }
        }
    }
}

void
SourceModifier::generate_usings(ProgElems *modified_decl, const MultiMethod *multimethods)
{
  const string &meth_name = multimethods->getMethodName();
  const string meth_name_changed = change_method_name (meth_name);
  const InheritFromBases &bases = multimethods->getInheritFromBases();

  for (InheritFromBasesIterator bit = bases.begin ();
        bit != bases.end (); ++bit)
    {
      modified_decl->add_item
          (new NormalText ("using " +
            *bit + "::" +
            meth_name_changed + ";\n"));
    }
}

void
SourceModifier::transform_class_decl(ClassDecl *cl, ProgElems *modified_source)
{
  const string class_name = cl->getClassName ();
  bool do_generate_line_info = ! should_invade_target (class_name) &&
    (! (clean_targets && ! cl->contains_multimethods()));

  // first we generate the forward declarations
  GeneratedForwardDecls forward_decls;

  const MultiMethods *multimethods = cl->getMultiMethods ();
  //const Methods *methods = 0;

  if (multimethods)
    {
      for (MultiMethodTypeIterator bit = multi_method_types.begin(cl);
           bit != multi_method_types.end(cl); ++bit)
        {
          fill_forward_decl_param(*bit, forward_decls);
        }
    }

  fill_target_forward_decls(cl, forward_decls);

  if (forward_decls.size () > 0)
    {
      // generate forward declarations of classes used in double
      // dispatch methods
      modified_source->add_item
        (new NormalText (generate_comment_tag (FORWARD_DECL)));

      for (GeneratedForwardDecls::const_iterator fit = forward_decls.begin ();
           fit != forward_decls.end (); ++fit)
        {
          const string &className = (*fit)->getClassName();
          modified_source->add_item
            (new NormalText ((*fit)->getClassStruct() + " " + className + ";" +
                             generate_original_file_info (className) + "\n"));
        }

      modified_source->add_item
        (new NormalText (generate_comment_tag (END_TAG)));

      modified_source->add_item (new NormalText ("\n"));
    }

  if (do_generate_line_info)
    insert_line_info (modified_source, class_name, cl);

  ostringstream cl_decl;

  cl_decl << cl->getClassStruct() << " " << class_name;

  const BaseClasses *base_classes = cl->getBaseClasses ();

  if (base_classes)
    {
      cl_decl << " : ";
      BaseClass *base;
      for (BaseClassesIterator bit = base_classes->begin ();
           bit != base_classes->end ();)
        {
          base = *bit;
          cl_decl << base->derivToString () << " " <<
                  (base->isVirtual () ? "virtual " : "") <<
                  base->getClassName();

          if (++bit != base_classes->end ())
            cl_decl << ", ";
        }
    }

  modified_source->add_item
    (new NormalText (cl_decl.str () + "\n"));
  modified_source->add_item
    (new NormalText ("{"));
}

void
SourceModifier::transform_includeheader(IncludeHeader *ih, ProgElems *source)
{
  /*source->add_item
    (new NormalText
     ("#include " + ih->getText ()));*/
  source->add_item (new IncludeHeader (*ih));
}

const string
SourceModifier::generate_original_file_info(const string &name)
{
  ProgElems *source = sourcestructs->retrieve_source (name);

  if (source)
    return " // file: " + source->getFileName();

  return "";
}

void
SourceModifier::fill_forward_decl_param(const BranchType &btype, GeneratedForwardDecls &forward_decls, int pos)
{
  const DeclType *decl_type;;
  const MethParams *params;
  int counter = 1;

  params = btype.getParams();
  for (MethParamsIterator pit = params->begin ();
      pit != params->end (); ++pit, ++counter)
  {
    decl_type = (*pit)->getDeclType ();
    if (counter >= pos && ! DeclType::isBaseType (decl_type))
      forward_decls.insert (decl_type->getClassDecl());
  }
  decl_type = btype.getRetType ();
  if (! DeclType::isBaseType (decl_type))
    forward_decls.insert (decl_type->getClassDecl());
}

// TODO: remove when new alg is implemented?
const string
SourceModifier::dispatch_method_name(const MethDecl *meth)
{
  return dispatch_method_name(meth->getMethodName(), meth->getSpecializingParam());
}

const string
SourceModifier::dispatch_method_name(const string &name, const MethParam *spec_param)
{
  string method_name = name;

  if (spec_param->isConst())
    method_name += "_const";

  return "dispatch_" + method_name;
}

const string
SourceModifier::change_method_name(const MethDecl *meth)
{
  return change_method_name (meth->getMethodName());
}

const string
SourceModifier::change_method_name(const string &meth)
{
  if (rename_overloaded)
    return meth + renamed_suffix;
  else
    return meth;
}

const string
SourceModifier::change_main_method_name(const string &meth)
{
  if (! rename_overloaded)
    return meth + renamed_main_suffix;
  else
    return meth;
}

const string
SourceModifier::generate_comment_tag(const string &s)
{
  return COMMENT_TAG + s + ", DO NOT MODIFY\n";
}

// generate a "#line <line_number> <file_name>
const string
SourceModifier::generate_line_info(const string &name,
                                   const WithTokenInfo *p,
                                   bool last)
{
  ProgElems *orig_source = sourcestructs->retrieve_source (name);

  assert (orig_source);

  ostringstream s;

  s << "#line " << (last ? p->getLastLine() : p->getLine()) << " \"" << main_path
    << orig_source->getFileName () << "\"\n";

  return s.str();
}

// insert a "#line <line_number> <file_name>
void
SourceModifier::insert_line_info(ProgElems *source, const string &name,
                                 const WithTokenInfo *p, bool last)
{
  if (no_linenum)
    return;

  source->add_item
    (new NormalText (generate_line_info (name, p, last)));
}


