/***************************************************************************
 *   Copyright (C) 2003 by Lorenzo Bettini                                 *
 *   http://www.lorenzobettini.it                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "sourcemodifierdouble.h"

#include "classend.h"
#include "sourcestructs.h"
#include "my_sstream.h"
#include "forward_method.h"
#include "multimethodtypes.h"
#include "multimethodtype.h"
#include "mainverbose.h"
#include "normaltext.h"
#include "methparams.h"

#include <assert.h>

#define FORWARDMETH_PREFIX "_forward_"
#define CLIENT_OBJECT "_o"

using std::ostringstream;

static const string
const_cast_obj(const MethParam *param, const string &t, const string &obj_name)
{
  if (param->isConst())
    return "(const_cast<" + t + ">(" + obj_name + "))";

  return obj_name;
}

static const string
cast_to_const_obj(const MethParam *param, const string &t, const string &obj_name)
{
  if (param->isConst())
    return "(const " + t + ")(" + obj_name + ")";

  return obj_name;
}

void
SourceModifierDouble::generate_main_method(ProgElems *modified_meth,
                                    const BranchType &bt)
{
  const MethParam *specializing_param = bt.getParam();
  const string &top_spec_arg = specializing_param->getTypeName();
  const string &method_name = bt.getMethodName();

  generate_include_header
    (top_spec_arg, modified_meth,
     should_use_output_extension(top_spec_arg));

  main_method_gen_class main_meth_gen;
  ostringstream main_meth_string;

  fill_main_method_generator (main_meth_gen, bt);

  const string param_name =
    const_cast_obj
      (specializing_param, top_spec_arg + " *",
        specializing_param->getText());

  const string &ret_type = bt.getRetType()->toString();
  const MethParams *params = bt.getParams();

  main_meth_gen.set_body
    (string((ret_type != "void" ? "return " : "")) +
     param_name +
     "->" + dispatch_method_name (method_name, specializing_param) +
     "(this" +
     (params->size () > 1 ? ", " + params->toStringParamNames (2) :
      "") +
     ");");

  main_meth_gen.generate_main_method (main_meth_string);

  modified_meth->add_item
    (new NormalText (main_meth_string.str ()));
}

void
SourceModifierDouble::transform_classend(ClassEnd *classend,
                                   ProgElems *modified_source,
                                   ProgElems *modified_meth)
{
  // right before the end of the class declaration, we have to add
  // the declaration of dispatch methods

  const ClassDecl *cl = classend->getClassDecl();
  const string &class_name = cl->getClassName();

  verbose->print("disp_DB methods in class " + class_name);
  verbose->print(extended_types.toStringParam(cl));

  bool is_to_invade = should_invade_target (class_name);
  bool found_one = false;
  bool is_first = true;

  main_method_gen_class disp_meth_gen;
  disp_meth_gen.set_class_name (class_name);

  for (MultiMethodTypeIterator bit = extended_types.begin_Param(cl); bit != extended_types.end_Param(cl); ++bit) {
    if (is_first) {
      modified_source->add_item
        (new NormalText (generate_comment_tag (DISPATCH_METHS)));
      modified_source->add_item (new NormalText ("public:\n"));
      is_first = false;
    }

    found_one = true;

    const BranchType &branch_type = *bit;
    const ClassDecl *Specializing_class = branch_type.getReceiverType();
    const string specializing_class = Specializing_class->getClassName();
    const string return_type = branch_type.getRetType()->toString();
    const MethParam *spec_param = branch_type.getParam();
    const string &method_name = branch_type.getMethodName();
    const string dispatch_meth = dispatch_method_name (method_name, spec_param);
    const MethParams *params = branch_type.getParams ();

    const string other_params =
      (params->size () > 1 ? ", " + params->toString (2) : "");

    const string signature =
      return_type + " " +
      dispatch_meth + "(" +
      specializing_class + " *" + other_params + ")";

    modified_source->add_item
    (new NormalText
      ("virtual " + signature + ";\n" ));

    if (is_to_invade)
      {
        modified_meth = modified_methods->retrieve_source(specializing_class);
        assert (modified_meth);
        generate_include_header (class_name, modified_meth, false);
      }
    else
      generate_include_header (specializing_class, modified_meth);

    ostringstream disp_meth_buff;

    bool use_forward = (branch_type.getAccessType() != PUBLIC);

    string client_obj = specializing_class + CLIENT_OBJECT;

    disp_meth_gen.set_ret_type (return_type);
    disp_meth_gen.set_meth_name (dispatch_meth);
    disp_meth_gen.set_param_string
      (specializing_class + " *" +
      client_obj +
      other_params);
    disp_meth_gen.set_body
      (string((return_type != "void" ? "return " : "")) +
      client_obj + "->" +
      (use_forward ? FORWARDMETH_PREFIX : "") +
      change_method_name (method_name) +
      "(" + cast_to_const_obj(spec_param, class_name + " *", "this") +
      (params->size () > 1 ?
        ", " + params->toStringParamNames (2) :
        "") +
      ");");
    disp_meth_gen.generate_main_method (disp_meth_buff);

    modified_meth->add_item
      (new NormalText (disp_meth_buff.str ()));
  }

  if (found_one)
    modified_source->add_item
      (new NormalText (generate_comment_tag (END_TAG)));

  modified_source->add_item (new NormalText (classend->getText()));
}

void
SourceModifierDouble::handle_restricted_methods(ProgElems *modified_decl, const string &meth, const ClassDecl *cl)
{
  modified_decl->add_item (new NormalText ("public:\n"));
  AccessType previous_access_type = PUBLIC;

  for (MultiMethodTypeIterator bit = extended_types.begin(meth, cl);
       bit != extended_types.end(meth, cl); ++bit) {
    const BranchType &branch_type = *bit;

    if (branch_type.getAccessType() == PUBLIC)
      continue;

    previous_access_type = branch_type.getAccessType();

    const ClassDecl *Specializing_class = branch_type.getReceiverType();
    const string specializing_class = Specializing_class->getClassName();
    const string return_type = branch_type.getRetType()->toString();
    const MethParams *params = branch_type.getParams ();

    forward_method_gen_class forward_method;
    ostringstream forward_method_str;

    forward_method.set_ret_type (return_type);
    forward_method.set_meth_name (FORWARDMETH_PREFIX + change_method_name (meth));
    forward_method.set_param_string (params->toString ());
    forward_method.set_body
      (string((return_type != "void" ? "return " : "")) +
              change_method_name (meth) + "(" +
              params->toStringParamNames () +
              ");");

    forward_method.generate_forward_method (forward_method_str);

    modified_decl->add_item
      (new NormalText (forward_method_str.str ()));
  }

  assert (previous_access_type != PUBLIC);

  modified_decl->add_item
   (new NormalText (translate_access_type (previous_access_type)+"\n"));
}

void
SourceModifierDouble::fill_target_forward_decls(const ClassDecl *cl,
    GeneratedForwardDecls &forward_decls)
{
  if (clean_targets)
    return;

  for (MultiMethodTypeIterator bit = extended_types.begin_Param(cl); bit != extended_types.end_Param(cl); ++bit)
    {
      forward_decls.insert (bit->getReceiverType());
      fill_forward_decl_param(*bit, forward_decls, 2);
    }
}

